%       $Id: makexdy.w,v 1.5 2004/03/20 17:18:35 bronger Exp $    

@q======================================================================@>
@q    makexdy.w - Part of the tbook Program                             @>
@q    Copyright 2002--2004 Torsten Bronger                              @>
@q                         <bronger@@users.sourceforge.net>             @>
@q                                                                      @>
@q  This program is free software; you can redistribute it and/or       @>
@q  modify it under the terms of the Artistic License 2.0 as published  @>
@q  by Larry Wall.  You should have received a copy of the Artistic     @>
@q  License 2.0 along with this program in the file COPYING; if not,    @>
@q  you can get it at                                                   @>
@q    http://dev.perl.org/rfc/346.html                                  @>
@q  or contact the current maintainer of tbook.                         @>
@q                                                                      @>
@q  This program is distributed in the hope that it will be useful,     @>
@q  but WITHOUT ANY WARRANTY; without even the implied warranty of      @>
@q  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       @>
@q  Artistic License 2.0 for more details.                              @>
@q                                                                      @>
@q  This file may only be distributed together with a copy of the       @>
@q  tbook program.                                                      @>
@q                                                                      @>
@q  The tbook program consists of all files listed in manifest.txt.     @>
@q                                                                      @>
@q                                                                      @>
@q  To get a compilable C++ source code, use                            @>
@q       ctangle makexdy.w - makexdy.cc                                 @>
@q  To get a structured documentation of the program, use               @>
@q       cweave makexdy.w                                               @>
@q       tex makexdy                                                    @>
@q                                                                      @>
@q  cweave and ctangle belong to the cweb system by Levy/Knuth.         @>
@q  Further information can be obtained at                              @>
@q       http://www-cs-faculty.stanford.edu/~knuth/cweb.html            @>
@q======================================================================@>

\secpagedepth=1
\def\TeX{T\kern-.1667em\lower.5ex\hbox{E}\kern-.125emX}
\def\LaTeX{L\kern-.36em%
        {\setbox0=\hbox{T}%
         \vbox to\ht0{\hbox{\sevenrm A}\vss}%
        }%
        \kern-.15em%
        \TeX}
\def\LaTeXbf{L\kern-.36em%
        {\setbox0=\hbox{T}%
         \vbox to\ht0{\hbox{\sevenbf A}\vss}%
        }%
        \kern-.15em%
        \TeX}
\def\AMSinner{\cal A\mkern-2mu\raise-0.5ex\hbox{$\cal M$}\mkern-2muS}
\def\AMS{\ifmmode\AMSinner\else$\AMSinner$\fi}
\font\sf=cmss10 \font\sfbf=cmssbx10
\font\sfa=cmss7
\def\tbook{{\sfbf t\sf book}}

\font\stitlefont=cmss8 scaled 1815         % sans serif type in title
\font\sbtitlefont=cmssbx10 scaled\magstep2 % sans bold type in title
\font\ttitlefont=cmtt8 scaled 1815         % typewriter type in title


\def\title{MAKEXDY (Version 1.5)}
\def\topofcontents{\null\vfill
  \centerline{\titlefont The {\sbtitlefont t\stitlefont book} index sorting
  creator {\ttitlefont MAKEXDY}}
  \vskip 15pt
  \centerline{(Version 1.5)}
  \vfill}
\def\botofcontents{\vfill
\noindent
Copyright \copyright\ 2002--2004 Torsten Bronger
                           ({\tt bronger@@users.sourceforge.net})
\bigskip\noindent
Permission is granted to make and distribute verbatim copies of this
document provided that the copyright notice and this permission notice
are preserved on all copies.

\smallskip\noindent
Permission is granted to copy and distribute modified versions of this
document under the conditions for verbatim copying, provided that the
entire resulting derived work is given a different name and distributed
under the terms of a permission notice identical to this one.
}

@i c++lib.w
@s ios int

@** Introduction.  This program \.{makexdy} (``make xdy file'') is to be used
as part of the \tbook\ system.  It takes a filename as the only argument and
creates a xindy style file fragment.  This fragment contains the lettergroups
and merge rules for correct sorting when compiling the index, and it uses
\hbox{{\mc UTF}-8} encoding.

@c
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <stdio.h>
#include <cstring>
#include <typeinfo>

using namespace std;

@** Wide characters with |xchar|.  The first thing I need is a way to handle
unicodes in a convenient way.  For this I define a data type called |xchar|
that is by and large a wide character.  Unfortunately my current {\mc
GNU}~\CPLUSPLUS/ compiler is rather weak in this point.  Therefore I cannot use
wide strings for example which would make the whole thing more
homogeneous. {\mc GNU}~\CPLUSPLUS/ version~3 should not have these problems
anymore.

@s xchar int

@c
class xchar;

typedef enum { enc_latin1, enc_utf8 } encodingT;

@* Streams for flexible encoding.  I want to be able to read and write |xchar|s
in a simple way via streams.  The problem is that I wanted to be able to set
the encoding with a manipulator.  I don't know why anymore.  But be that as is
may, so |xifstream| and |xofstream| came into existence.

@s xifstream int
@s xofstream int

@c
class xifstream : public ifstream {
    encodingT encoding;
    friend xifstream& operator>>(xifstream &in, xchar &ch);
public:
    xifstream(const string s) : encoding(enc_latin1), ifstream(s.c_str()) { }
};

class xofstream : public ofstream {
    encodingT encoding;
    friend xofstream& operator<<(xofstream &out, const xchar &ch);
    friend xofstream& utf8(xofstream& s);
    friend xofstream& latin1(xofstream& s);
public:
    xofstream(const string s) : encoding(enc_latin1), ofstream(s.c_str()) { }
    xofstream& operator<<(xofstream& (*f)(xofstream&)) { return f(*this); }
};

@ These are the two above mentioned manipulators.  I can choose between
\hbox{{\mc UTF}-8} and \hbox{latin-1}.

@c
xofstream& utf8(xofstream& s) {
    s.encoding = enc_utf8;
    return s;
}

xofstream& latin1(xofstream& s) {
    s.encoding = enc_latin1;
    return s;
}

@* Definition of |xchar|. Now for the |xchar| class itself.  Its unicode number
is in |contents|.  It is by and large trivial.

@c
class xchar {
    wchar_t contents;
    friend string xindy_armor(xchar w, const bool ascii_armor = true);
    friend void write_as_utf8(ostream& out, const xchar ch);
    friend void read_as_utf8(istream& in, xchar& ch);
    friend xchar read_as_utf8(istream& in);
    friend xofstream& operator<<(xofstream &out, const xchar &ch);
    friend xifstream& operator>>(xifstream &in, xchar &ch);
    operator int() const { return int(contents); } // forbidden
    operator char() const { return char(contents); } // forbidden
public:
    xchar() @+ : contents(0xffff) { }
    xchar(wchar_t u) @+ { @+ contents = u; @+ }
    xchar(const xchar& ch) @+ : contents(ch.contents) { }
    xchar& operator=(const xchar& ch) { contents = ch.contents; return *this; }
    operator wchar_t() const @+ { @+ return contents; @+ }
    xchar& operator=(const long int value) { contents = value; return *this; }
    xchar& operator=(const wchar_t value) { contents = value; return *this; }
};

@** I/O with \hbox{{\mc UTF}-8}.

@* Output.  |shift_and_mask()| is a helper routine for |write_as_utf8|.  The
latter actually writes an |xchar| into a stream in \hbox{{\mc utf}-8} encoding.

@c
inline char shift_and_mask(long int value, int shift_value = 0,
			   unsigned int mask = 0x3f,
			   unsigned int prefix = 0x80) {
    return char((((unsigned long)value >> shift_value) & mask) + prefix);
}

void write_as_utf8(ostream& out, const xchar ch) {
    if (ch.contents <= 0x7f) out.put(char(ch.contents));
    else if (ch.contents <= 0x7ff) {
	out.put(shift_and_mask(ch.contents,6,0x1f,0xc0));
	out.put(shift_and_mask(ch.contents));
    } else if (ch.contents <= 0xffff) {
	out.put(shift_and_mask(ch.contents,12,0x0f,0xe0));
	out.put(shift_and_mask(ch.contents,6));
	out.put(shift_and_mask(ch.contents));
    } else if (ch.contents <= 0x1fffff) {
	out.put(shift_and_mask(ch.contents,18,0x07,0xf0));
	out.put(shift_and_mask(ch.contents,12));
	out.put(shift_and_mask(ch.contents,6));
	out.put(shift_and_mask(ch.contents));
    } else if (ch.contents <= 0x3ffffff) {
	out.put(shift_and_mask(ch.contents,24,0x03,0xf8));
	out.put(shift_and_mask(ch.contents,18));
	out.put(shift_and_mask(ch.contents,12));
	out.put(shift_and_mask(ch.contents,6));
	out.put(shift_and_mask(ch.contents));
    } else {
	out.put(shift_and_mask(ch.contents,30,0x01,0xfc));
	out.put(shift_and_mask(ch.contents,24));
	out.put(shift_and_mask(ch.contents,18));
	out.put(shift_and_mask(ch.contents,12));
	out.put(shift_and_mask(ch.contents,6));
	out.put(shift_and_mask(ch.contents));
    }
}

@* Input.  The following two routines do the same thing but in the other
direction.

@c
inline wchar_t mask_and_shift(char c, unsigned char mask, 
                                        int shift_value = 0) {
    return (unsigned long)((unsigned char)c & mask) << shift_value;
}

void read_as_utf8(istream& in, xchar& ch) {
    char c1;
    in.get(c1);
    unsigned int value = (unsigned int)c1;
    if (value < 128)
	ch.contents = value;
    else if ((value & 0xe0) == 0xc0) {  /* 11 bits in 2 bytes: first 2048
					   positions, umlauts etc. */
	char c2;
	in.get(c2);
	ch.contents = mask_and_shift(c1,0x1f,6) + mask_and_shift(c2,0x3f);
    } else if ((value & 0xf0) == 0xe0) {  /* 16 bis in 3 bytes: first 65536
					     positions, MathML, Indian etc. */
	char c2, c3;
	in.get(c2); in.get(c3);
	ch.contents = mask_and_shift(c1,0x0f,12) + 
	    mask_and_shift(c2,0x3f,6) + 
	    mask_and_shift(c3,0x3f);
    } else if ((value & 0xf8) == 0xf0) { /* 21 bits in 4 bytes: first 2
					    million positions, for MathML */
	char c2, c3, c4;
	in.get(c2); in.get(c3); in.get(c4);
	ch.contents = mask_and_shift(c1,0x07,18) + 
	    mask_and_shift(c2,0x3f,12) + 
	    mask_and_shift(c3,0x3f,6) + 
	    mask_and_shift(c4,0x3f);
    } else if ((value & 0xfc) == 0xf8) { /* Ultra Unicode */
	char c2, c3, c4, c5;
	in.get(c2); in.get(c3); in.get(c4); in.get(c5);
	ch.contents = mask_and_shift(c1,0x03,24) + 
	    mask_and_shift(c2,0x3f,18) + 
	    mask_and_shift(c3,0x3f,12) + 
	    mask_and_shift(c4,0x3f,6) +
	    mask_and_shift(c5,0x3f);
    } else if ((value & 0xfe) == 0xfc) { /* Ridiculous Unicode */
	char c2, c3, c4, c5, c6;
	in.get(c2); in.get(c3); in.get(c4); in.get(c5); in.get(c6);
	ch.contents = mask_and_shift(c1,0x01,30) +
	    mask_and_shift(c2,0x07,24) + 
	    mask_and_shift(c3,0x3f,18) + 
	    mask_and_shift(c4,0x3f,12) + 
	    mask_and_shift(c5,0x3f,6) +
	    mask_and_shift(c6,0x3f);
    } else /* Invalid code */ ;
}

@* I/O for |xchar|.  Here I use the above work to write the actual stream
operators for |xchar|, for both normal and \hbox{``x-''}streams.

@c
xofstream& operator<<(xofstream &out, const xchar &ch) {
    if (out.encoding == enc_utf8)
	write_as_utf8(out,ch);
    else {   // iso-8859-1
	if (ch.contents <= 0xff) out.put(char(ch.contents));
	else out.put('?');
    }
    return out;
}

ostream& operator<<(ostream &out, const xchar &ch) {
    write_as_utf8(out,ch);
    return out;
}

xifstream& operator>>(xifstream &in, xchar &ch) {
    if (in.encoding == enc_utf8)
	read_as_utf8(in,ch);
    else {  // iso-8859-1
	char c1;
	in.get(c1);
	unsigned int value = (unsigned int)c1;
 	ch.contents = value;
    }
    return in;
}

istream& operator>>(istream& in, xchar& ch) {
    read_as_utf8(in,ch);
    return in;
}

@** xindy routines.  Here I map character code numbers ({\it not\/} unicodes)
on Latin triplets.  So the lowest code gets |"AAA"|, the next |"AAB"|, until
|"zzz"|.  The idea behind that is that words transformed to such triplet
sequences can easily be sorted by xindy.

@c
string unicode_to_triplet(wchar_t unicode) {
    const char digits[52] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J',
			     'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
			     'U', 'V', 'W', 'X', 'Y', 'Z', 'a', 'b', 'c', 'd',
			     'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
			     'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
			     'y', 'z'};
#if 1==0
    /* This test is unnecessary since |wchar_t| is unfortunately only 
       2~bytes wide. */
    if (unicode < 0 || unicode >= 52*52*52) return string("???");
#endif
    int d0,d1,d2;
    d2 = unicode / (52*52);  unicode -= d2 * 52*52;
    d1 = unicode / 52;  unicode -= d1 * 52;
    d0 = unicode;
    string result;
    result = digits[d2];
    result += digits[d1];
    result += digits[d0];
    return result;
}

@* {\mc ASCII} armor for xindy.  This used to be interesting when I thought
that you could avoid all non-{\mc ASCII} characters in xindy style files, using
those nasty \.{\#\BS Code} commands.  Now it's almost absolete, and
|ascii_armor| is always |false| when calling this routine.

@c
string xindy_armor(xchar w, const bool ascii_armor) {
    unsigned int value = w.contents;
    string result;
    if (value < 128) {
	result = '"';
	result += char(value);
	result += '"';
    } else {
	stringstream ss;
	ss << w;
	if (!ascii_armor) return string("\"") + ss.str() + "\"";
	string utf8(ss.str());
	result = "(concatenate 'string";
	for (int i = 0; i < utf8.length(); i++) {
	    stringstream ss;
	    ss << " (string #\\Code" << int((unsigned char)(utf8[i])) << ")";
	    string xindy_format(ss.str());
	    for (int j = 0; j < xindy_format.length(); j++)
		result += xindy_format[j];
	}
	result += ')';
    }
    return result;
}

@* xindy rules creation.  Actually this is the real main routine.  It reads
|order| (which is mostly the file \.{utf8.ord}) and writes to standard out.
The input file format is very straightforward: Every line contains one letter
group.  In the very first column is the name (i.\,e.\ the main letter) of the
letter group, and all letters that follow belong to it, separated by {\it
one\/} space character each.  If two letters are not separated by a space, they
are treated totally the same in the sorting process.  This is typically the
case for upper/lower case letters.  So a typical line in \.{utf8.ord} may be:

{\tentt Aa \"A\"a \'A\'a \`A\`a \AA\aa}

Of course, the lines (letter groups) must have proper order within the file,
too.  Everything after a semicolon is a comment and ignored till the end of the
line.  Empty lines don't do any harm, but should be avoided with the current
implementation.  The file must use the \hbox{{\mc UTF}-8} encoding.

This routine writes the letter group definitions and all merge rules necessary
to do the sorting.  Further sort rules are unnecessary.

@c
void create_rules(istream& order) {
    ostringstream lettergroups, mergerules;
    xchar last_lettergroup_name(L'\0');
    lettergroups << "(define-letter-group \"sssymbols\")\n";
    bool beginning(true), new_prefix(false);
    int index = 0;
    xchar ch;
    while (order >> ch) {
	if (wchar_t(ch) == L'\n') {
	    index++;
	    beginning = true;
	    new_prefix = false;
	    continue; 
	}
	if (wchar_t(ch) == L' ' || wchar_t(ch) == L'\t') {
	    index++;
	    new_prefix = true;
	    continue;
	}
	if (wchar_t(ch) == L'\\') {
	    long code;
	    char prefix[3];
	    order.get(prefix,3);
	    if (string(prefix) != "0x") {
		cerr << "makexdy: Invalid hex code!" << endl;
		continue;
	    }
	    order >> hex >> code;
	    ch = code;
	}
	string prefix = unicode_to_triplet(index);
	if (new_prefix) {
	    new_prefix = false;
	    lettergroups << " \"" << prefix[0] << prefix[1] << prefix[2] << '"';
	}
	if (beginning) {
	    beginning = false;
	    if (wchar_t(ch) == L';') {  // skip comment
		do order >> ch; while(order && wchar_t(ch) != L'\n');
		beginning = true;
		continue; 
	    }
	    if (wchar_t(last_lettergroup_name) != L'\0')
		lettergroups << "))\n";
	    lettergroups << "(define-letter-group " << xindy_armor(ch,false);
	    lettergroups << " :after ";
	    if (wchar_t(last_lettergroup_name) != L'\0')
		lettergroups << xindy_armor(last_lettergroup_name,false);
	    else lettergroups << "\"sssymbols\"";
	    lettergroups << " :prefixes (" 
			 << '"' << prefix[0] << prefix[1] << prefix[2] << '"';
	    last_lettergroup_name = ch;
	}
	mergerules << "(merge-rule "
		   << xindy_armor(ch,false)
		   << " \""
		   << prefix[0] << prefix[1] << prefix[2]
		   << "\" :string)\n";
    }
    if (wchar_t(last_lettergroup_name) != L'\0') lettergroups << "))\n";
    lettergroups << "(define-letter-group \"ppparameters\" :after ";
    if (wchar_t(last_lettergroup_name) != L'\0')
	lettergroups << xindy_armor(last_lettergroup_name,false);
    else lettergroups << "\"sssymbols\"";
    lettergroups << ")\n\n";
    cout << "; -*- Mode: Lisp; coding: utf-8; -*-\n\n";
    cout << lettergroups.str() << mergerules.str();
}

@** The |main()| routine.  Nothing exciting happens here.  I have to be quite
pedantic as far as proper arguments are concerned becaus I must return a vaild
return value to the calling make process (or whatever).

@c
int main(int argc, char *argv[]) {
    if (argc < 2) {
	cerr << "makexdy: no filename given" << endl;
	exit(1);
    }
    if (argc > 2) {
	cerr << "makexdy: too many parameters" << endl;
	exit(1);
    }
    ifstream order(argv[1]);
    if (!order.good()) {
	cerr << "makexdy: file " << argv[1] << " not found" << endl;
	exit(1);
    }
    create_rules(order);
    return 0;
}

@** Index.
