%       $Id: tbcrent.w,v 1.17 2004/03/20 17:18:35 bronger Exp $

@q======================================================================@>
@q    tbcrent.w - Part of the tbook Program                             @>
@q    Copyright 2002--2004 Torsten Bronger                              @>
@q                         <bronger@@users.sourceforge.net>             @>
@q                                                                      @>
@q  This program is free software; you can redistribute it and/or       @>
@q  modify it under the terms of the Artistic License 2.0 as published  @>
@q  by Larry Wall.  You should have received a copy of the Artistic     @>
@q  License 2.0 along with this program in the file COPYING; if not,    @>
@q  you can get it at                                                   @>
@q    http://dev.perl.org/rfc/346.html                                  @>
@q  or contact the current maintainer of tbook.                         @>
@q                                                                      @>
@q  This program is distributed in the hope that it will be useful,     @>
@q  but WITHOUT ANY WARRANTY; without even the implied warranty of      @>
@q  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       @>
@q  Artistic License 2.0 for more details.                              @>
@q                                                                      @>
@q  This file may only be distributed together with a copy of the       @>
@q  tbook program.                                                      @>
@q                                                                      @>
@q  The tbook program consists of all files listed in manifest.txt.     @>
@q                                                                      @>
@q                                                                      @>
@q  To get a compilable C++ source code, use                            @>
@q       ctangle tbcrent.w - tbcrent.cc                                 @>
@q  To get a structured documentation of the program, use               @>
@q       cweave tbcrent.w                                               @>
@q       tex tbcrent                                                    @>
@q                                                                      @>
@q  cweave and ctangle belong to the cweb system by Levy/Knuth.         @>
@q  Further information can be obtained at                              @>
@q       http://www-cs-faculty.stanford.edu/~knuth/cweb.html            @>
@q======================================================================@>

\secpagedepth=1
\def\TeX{T\kern-.1667em\lower.5ex\hbox{E}\kern-.125emX}
\def\LaTeX{L\kern-.36em%
        {\setbox0=\hbox{T}%
         \vbox to\ht0{\hbox{\sevenrm A}\vss}%
        }%
        \kern-.15em%
        \TeX}
\def\LaTeXbf{L\kern-.36em%
        {\setbox0=\hbox{T}%
         \vbox to\ht0{\hbox{\sevenbf A}\vss}%
        }%
        \kern-.15em%
        \TeX}
\font\sf=cmss10 \font\sfbf=cmssbx10
\font\sfa=cmss7
\def\tbook{{\sfbf t\sf book}}

\font\stitlefont=cmss8 scaled 1815         % sans serif type in title
\font\sbtitlefont=cmssbx10 scaled\magstep2 % sans bold type in title
\font\ttitlefont=cmtt8 scaled 1815         % typewriter type in title

\def\title{TBCRENT (Version 1.5)}
\def\topofcontents{\null\vfill
  \centerline{\titlefont The {\sbtitlefont t\stitlefont book} Entities File
    Creator {\ttitlefont TBCRENT}}
  \vskip 15pt
  \centerline{(Version 1.5)}
  \vfill}
\def\botofcontents{\vfill
\noindent
Copyright \copyright\ 2002--2004 Torsten Bronger
                           ({\tt bronger@@users.sourceforge.net})
\bigskip\noindent
Permission is granted to make and distribute verbatim copies of this
document provided that the copyright notice and this permission notice
are preserved on all copies.

\smallskip\noindent
Permission is granted to copy and distribute modified versions of this
document under the conditions for verbatim copying, provided that the
entire resulting derived work is given a different name and distributed
under the terms of a permission notice identical to this one.
}


@i c++lib.w
@s ios int
@s line normal_identifier

@* Introduction.  This trivial program \.{tbcrent} (``\tbook\ create entities file'') is
to be used as part of the \tbook\ system.  It reads a lot of standard entity files
(extension ``\.{.ent}''), mostly from MathML, the rest from {\mc HTML}, and unites them in
one file, \.{tbents.txt}, which is then used by \.{tbrplent}.

@* The main Routine.  It consists of three parts: First, all entity files are defined and
connected with a file name.  Secondly, all entity files are read.  Thirdly, the entities
file is written.  Its file name is fixed, defined in |entities_file_name|.

@d entities_file_name "tbents.txt"

@<The main() function@>=
int main() {
    entities_by_unicode unicode_list;
    entities_by_names name_list;

    @<Define entity files@>@;
    @<Read entity files@>@;
    @<Mapping from |name_list| to |unicode_list|@>@;

    prefixed_ofstream tbents_txt(entities_file_name);
    write_entity_file(tbents_txt, unicode_list);
    return 0;
}

@ This is the list of entity files that are merged into the one entities file.  Most come
from MathML\spacefactor1000. However, \.{"tblatex.ent"} contains the \LaTeX\ macros.

@d number_of_files 37

@<Define entity files@>=
    char* filenames[number_of_files]={ "mathml/mmlalias.ent",
                       "iso8879/isoamsr.ent", "iso9573-13/isoamsr.ent",
                       "iso8879/isogrk1.ent",
                       "iso8879/isogrk2.ent",
                       "iso8879/isogrk4.ent", "iso9573-13/isogrk4.ent",
                       "iso8879/isogrk3.ent", "iso9573-13/isogrk3.ent",
                       "iso8879/isodia.ent", 
                       "html/lat1.ent", "html/special.ent", "html/symbol.ent",
                       "mathml/mmlextra.ent", 
                       "iso8879/isoamsa.ent", "iso9573-13/isoamsa.ent",
                       "iso8879/isoamsb.ent", "iso9573-13/isoamsb.ent",
                       "iso8879/isoamsc.ent", "iso9573-13/isoamsc.ent",
                       "iso8879/isoamsn.ent", "iso9573-13/isoamsn.ent",
                       "iso8879/isoamso.ent", "iso9573-13/isoamso.ent",
                       "iso8879/isobox.ent",
                       "iso8879/isocyr1.ent", "iso8879/isocyr2.ent",
                       "iso8879/isolat1.ent", "iso8879/isolat2.ent",
                       "iso9573-13/isomfrk.ent", 
                       "iso9573-13/isomopf.ent", "iso9573-13/isomscr.ent",
                       "iso8879/isonum.ent", "iso8879/isopub.ent",
                       "iso8879/isotech.ent", "iso9573-13/isotech.ent",
                       "tblatex.ent"
                       };

@ Here we finally read all entity files and add their contents to |unicode_list|.

@<Read entity files@>=
    for (int i = 0; i < number_of_files; i++) {
        string filename(filenames[i]);
        prefixed_ifstream entity_file(filename);
        read_entity_file(entity_file, name_list);
    }


@* Global Definitions.  Here is everything that should come first.  Since it's very
unconvenient to have these ubiquitous ``|std::@[@t@>@]|'' flying around here, I'm lazy and
make it implicit.

@c
@h
@<Header files inclusions@>@;

using namespace std;

@ The global constant |file_name_prefix| contains the directory that contains the file
\.{tbents.txt}.

@c
const char* tblibdir_env_raw = getenv("TBLIBDIR");
const string tblibdir_env = (tblibdir_env_raw == NULL ? "" : tblibdir_env_raw);
#ifdef TBLIBDIR
const string file_name_prefix(tblibdir_env.size() > 0 ? tblibdir_env : TBLIBDIR);
#else
const string file_name_prefix(tblibdir_env.size() > 0 ? tblibdir_env : "");
#endif

@* The Prefixed Files Stream Classes.  In order to be free as far as the directory where
the files are stored is concerned, I need a variant if the |ifstream| class that is aware
of |TBLIBDIR|.

Corresponding thing follows for |ofstream|.

@s prefixed_ifstream int
@s prefixed_ofstream int

@c
class prefixed_ifstream : public ifstream {
public:@/
    prefixed_ifstream(string filename) : ifstream() {
        @<Construct complete file name and open stream@>@;
    }    
}; @#


class prefixed_ofstream : public ofstream {
public:@/
    prefixed_ofstream(string filename) : ofstream() {
        @<Construct complete file name and open stream@>@;
    }    
};

@ Here we construct the complete (``prefixed'') filename, i.\,e.\ the filename with the
default directory given by |TBLIBDIR|, and open the stream in its default mode.

@<Construct complete file name and open stream@>=
        string prefixed_file_name(file_name_prefix);
        if (!file_name_prefix.empty()) {
            if (file_name_prefix[file_name_prefix.size()-1] != '/') // Ending slash lacking?
                prefixed_file_name = prefixed_file_name + "/";
        }
        prefixed_file_name = prefixed_file_name + filename;
        open(prefixed_file_name.c_str());
        if (!this->good()) cerr << "tbcrent: File " << prefixed_file_name << 
            " couldn't be opened" << endl;


@* Decimal Conversion.  We need helper routines to convert decimal to hexadecimal numbers
and vice versa.

For |string_to_dec()|, we first need |isalpha()| and things like that; in addition, the
functions are realised via string streams (wonderful thing).

@<Header...@>=
#include <cstdlib>
#include <sstream>
#include <iostream>

@ First the hexadecimal to decimal conversion.  The routine takes its argument |s|,
interprets it as a hexadecimal number like \.{0x7a} and returns its decimal value.

For hexadecimal conversion, the number must follow an \.{'x'}.  If it is not, the function
assumes that |s| is a decimal number and converts it as such.

@c
int string_to_dec(string number) {
    istringstream ss(number);
    int result = 0;

    if (number.find("#") != string::npos) return 0;
        /* A ``double'' entity? (Can't cope with that yet.) */
    if (number.find("x") == string::npos)  
        /* Is |number| decimal or hexadecimal? */
        ss >> result;
    else {
        char c;
        do ss >> c; while (c != 'x');
        ss >> hex >> result;
    } 
    return result;
}

@ Now for the other direction.  The function for the conversion becomes nearly trivial
with a string stream.  Actually, it's only a manipulator.

@c
string dec_to_hex(int a) {
    ostringstream result;

    result << "0x" << hex << a;
    return result.str();
}

@* The Class for Entities.  |unicode| is the unicode number.  |name| contains the name of
the entity (e.\,g.\ \.{ndash}).  |latex_macro| is a neat \LaTeX\ representation for that
entity, e.\,g.\ \.{\\copyright}.

It doesn't accept new values if it already has ones.  That mimics {\mc XML}'s behaviour
that the first definition counts.

@s entity int

@c
class entity {
    int unicode;
    string name, latex_macro;
    bool unicode_changed, name_changed, latex_macro_changed;
public:
    entity() : name_changed(false), latex_macro_changed(false),
               unicode_changed(false), unicode(0) { }
    void set_unicode(const int u);
    void set_name(const string n);
    void set_latex_macro(const string m);
    int get_unicode() const @+{@+ return unicode;@+ }
    string get_name() const @+{@+ return name;@+ }
    string get_latex_macro() const @+{@+ return latex_macro;@+ }
};

@ The three non-inline element function are trivial.

@c
void entity::set_unicode(const int u) {
    if (u == 0) return;
    if (!unicode_changed) {
        unicode = u;
        unicode_changed = true;
    } else if (unicode != u) 
               cerr << "tbcrent: Entity \"" << name << "\" with unicode "
                    << unicode << " is assigned unicode " << u << endl;
}

void entity::set_name(const string n) {
    if (n.empty()) return;
    if (!name_changed) {
        name = n;
        name_changed = true;
    }
}

void entity::set_latex_macro(const string m) {
    if (m.empty()) return;
    if (!latex_macro_changed && !m.empty()) {
        latex_macro = m;
        latex_macro_changed = true;
    } else if (latex_macro != m)
               cerr << "tbcrent: Entity \"" << name << "\" with LaTeX macro \""
                    << latex_macro << "\" is assigned LaTeX macro \"" << m 
                    << "\"" << endl;
}

@ We've needed |string|s here.

@<Header...@>=
#include <string>

@* The Containers for the Entity Table.  We also need other new datatypes, namely
|entities_by_unicode| and |entities_by_names|.  |entities_by_unicode| is an array of
|entity|s with a key that is the entity's unicode.  |entities_by_names| uses the {\it
name\/} (e.\,g.\ \.{"copy"} for the {\mc HTML} entity \.{"\&copy;"}) of the respective
entity as the key.

@s entities_by_unicode int
@s entities_by_names int

@c
typedef map<int,entity> entities_by_unicode;
typedef map<string,entity> entities_by_names;

@ We've needed a |map| here.

@<Header...@>=
#include <map> 


@* Reading Entity Files.  The entity file is read line by line.  If the line contains an
\.{"<!ENTITY "}, the routine expects an entity definition to follow.  |entity_name| stores
the entity name, which is then used as the key in the |unicode_list| map (because this is the
only thing that is {\it always\/} present).

The entity substitution (between double quotes) is scanned and stored in |subst|.  Now
there are three possibilities for |subst|: It may be a (possibly hexadecimal) number, then
it's stored in the |unicode_list| map, or it's a \LaTeX\ macro and thus stored in the local
|names| map.  

The third possibility is that it's entity name starts with \.{uc--}.  Then the following
number is interpreted as the unicode.  It's values are stored in |unicode_list|.  These are
the only entries there that are already finished.  The rest must be completed later.

This whole procedure can easily be broken (e.\,g.\ by two entity definition in {\it one\/}
line), but for now it's okay.  Moreover, \LaTeX\ macros in substitutions must never start
with \.{\&}, and must never be empty.  (Use \.{\{\}\&} and \.{\{\}} instead.)

@s size_type int

@c
void read_entity_file(ifstream& in, entities_by_names& name_list) {
    string line;
    string::size_type position;

    while (getline(in,line)) {
        position = line.find("<!ENTITY ");
        if (position != string::npos) {
            string entity_name, subst;
            position += 9;
            while (isspace(line[position])) position++;
            if (line[position] != '%') {  // {\it not\/} a parameter entity?
                while (!isspace(line[position])) entity_name += line[position++];
                while (line[position] != '\"') position++;
                position++;
                while (line[position] != '\"') subst += line[position++];
                if (subst[0] == '&') {
                    const int unicode = string_to_dec(subst.substr(2,subst.size()-3));
                    name_list[entity_name].set_name(entity_name);
                    name_list[entity_name].set_unicode(unicode);
                } else if (entity_name.substr(0,4) == "uc--") {
                    const int unicode = string_to_dec(entity_name.substr(4,entity_name.size()-4));
                    name_list[entity_name].set_name(entity_name);
                    name_list[entity_name].set_unicode(unicode);
                    name_list[entity_name].set_latex_macro(subst);
                } else name_list[entity_name].set_latex_macro(subst);
            }
        }
    }
}

@ We've needed file streams and |iswhite()| here.

@<Header...@>=
#include <fstream>
#include <cctype>

@ |unicode_list| is still empty.  That needn't be, because now everything is
available in |name_list|.  So here we step through the whole |name_list| map and get the
respective entries for |unicode_list| by using the entity names as key.

@<Mapping from |name_list| to |unicode_list|@>=
    typedef entities_by_names::const_iterator ci;
    for (ci p=name_list.begin(); p!=name_list.end(); ++p) {
        const int unicode = p->second.get_unicode();
        if (unicode != 0) {
            unicode_list[unicode].set_unicode(unicode);
            unicode_list[unicode].set_name(p->first);
            unicode_list[unicode].set_latex_macro(p->second.get_latex_macro());
        }
    }

@* Writing the Entities File.  The entities file has three columns: Unicode, entity name
and \LaTeX\ macro, the latter being enclosed by double quotes.

The first character in the line is an explicit \.{'0'}.  It's good for two reasons: First,
a hexadecimal number starts with \.{"x"} in {\mc XML} entity files, but should start with
\.{"0x"} (now mainly for aesthetical reasons, but there was a time when I hoped that a
\CPLUSPLUS/ input streams could read hexadecimal numbers).  Additionally it creates a
default number if |unicode| is empty.

In order to have nicer output, I insert void lines if the gap to the last unicode is
bigger than one.  For a gap of |100| and |1000| an additional black line is added.

@s ci int

@c
void write_entity_file(ostream& out, entities_by_unicode& unicode_list) {
    typedef entities_by_unicode::const_iterator ci;
    int last_unicode = 0;

    for (ci p=unicode_list.begin(); p!=unicode_list.end(); ++p) {
        const int unicode = p->first;
        const entity e = p->second;
        const string name = e.get_name(), latex_macro = e.get_latex_macro();

        if (latex_macro.empty()) continue;
        if (last_unicode != 0) {
            if (unicode - last_unicode > 1) out << '\n';
            if (unicode - last_unicode > 100) out << '\n';
            if (unicode - last_unicode > 1000) out << '\n';
        }
        out << "0x" << hex << unicode << '\t'  << name;
        for (int i=name.size(); i < 25; i++) out << ' ';
        out << '\"' << latex_macro << '\"' << '\n';
        last_unicode = unicode;
    }
}

@ We've used parameterised manipulators here.

@<Header...@>=
#include <iomanip>

@ Here finds the |main()| routine its home.

@c
@<The main() function@>

@* Index.
